/*
 * n01_stats
 *
 * csv.c
 *
 * Copyright (C) 1996-2006 by Nakashima Tomoaki. All rights reserved.
 *		http://www.nakka.com/
 *		nakka@nakka.com
 */

/* Include Files */
#define _INC_OLE
#include <windows.h>
#undef	_INC_OLE
#include <tchar.h>

#include "general.h"
#include "file.h"
#include "String.h"
#include "Memory.h"
#include "data.h"
#include "csv.h"

#include "resource.h"

/* Define */
#define FILELIST_ALLOC_COUNT	100

/* Global Variables */
// t@C
typedef struct _FILE_INFO {
	UINT file_name_hash;
	TCHAR file_name[FILE_NAME_SIZE];
} FILE_INFO;

static FILE_INFO *fli;
static int fli_count;
static int fli_size;

static BOOL cancel_flag;

extern OPTION op;

/* Local Function Prototypes */
static int __cdecl filelist_compare(const void *arg1, const void *arg2);
static BOOL filelist_add(FILE_INFO *new_fli);

static BOOL string_to_time(TCHAR *buf, SYSTEMTIME *tm);
static void get_statistics(const int type, DATA_INFO *di, const TCHAR *title, const TCHAR *st);
static TCHAR *csv_get_next(TCHAR *buf);
static TCHAR *csv_del_qoute(TCHAR *buf);
static void csv_get_data(DATA_INFO *di, TCHAR *buf);
static TCHAR *csv_put_conv(const TCHAR *buf);

/*
 * filelist_alloc - t@CXg̊m
 */
BOOL filelist_alloc(const int count)
{
	FILE_INFO *tmp_fli;

	if (count > fli_size) {
		// Ċm
		tmp_fli = mem_calloc(sizeof(FILE_INFO) * (count + FILELIST_ALLOC_COUNT));
		if (tmp_fli == NULL) {
			return FALSE;
		}
		if (fli != NULL) {
			CopyMemory(tmp_fli, fli, sizeof(FILE_INFO) * fli_count);
			mem_free(&fli);
		}
		fli = tmp_fli;
		fli_size = count + FILELIST_ALLOC_COUNT;
	}
	return TRUE;
}

/*
 * filelist_get_count - t@CXg̎擾
 */
int filelist_get_count(void)
{
	return fli_count;
}

/*
 * filelist_compare - t@CXg̔r
 */
static int __cdecl filelist_compare(const void *arg1, const void *arg2)
{
	FILE_INFO *fli1 = (FILE_INFO *)arg1;
	FILE_INFO *fli2 = (FILE_INFO *)arg2;

	if (fli1->file_name_hash < fli2->file_name_hash) {
		return -1;
	}
	if (fli1->file_name_hash > fli2->file_name_hash) {
		return 1;
	}
	return lstrcmpi(fli1->file_name, fli2->file_name);
}

/*
 * filelist_create - t@CXg̍쐬
 */
BOOL filelist_create(const DATA_INFO *di, const int di_count)
{
	TCHAR *name = TEXT("");
	UINT hash = 0;
	int i;

	mem_free(&fli);
	fli_size = di_count + FILELIST_ALLOC_COUNT;
	fli_count = 0;
	fli = mem_calloc(sizeof(FILE_INFO) * fli_size);
	if (fli == NULL) {
		return FALSE;
	}
	for (i = 0; i < di_count; i++) {
		if (hash == di[i].file_name_hash && lstrcmpi(name, di[i].file_name) == 0) {
			// A铯t@C̓XLbv
			continue;
		}
		hash = fli[fli_count].file_name_hash = di[i].file_name_hash;
		lstrcpy(fli[fli_count].file_name, di[i].file_name);
		name = fli[fli_count].file_name;
		fli_count++;
	}
	// בւ
	qsort(fli, fli_count, sizeof(FILE_INFO), filelist_compare);
	return TRUE;
}

/*
 * filelist_add - t@CXgɒǉ
 */
static BOOL filelist_add(FILE_INFO *new_fli)
{
	FILE_INFO *tmp_fli;
	int low, mid, high;

	// ǉʒǔ
	low = 0;
	high = fli_count - 1;
	while (low <= high) {
		mid = (high + low) / 2;
		switch (filelist_compare(&fli[mid], new_fli)) {
		case -1:
			low = mid + 1;
			break;
		case 1:
			high = mid - 1;
			break;
		case 0:
			// ɒǉĂ
			return FALSE;
		}
	}
	if (fli_count + 1 > fli_size) {
		// Ċm
		tmp_fli = mem_calloc(sizeof(FILE_INFO) * (fli_size + FILELIST_ALLOC_COUNT));
		if (tmp_fli == NULL) {
			return FALSE;
		}
		if (fli != NULL) {
			CopyMemory(tmp_fli, fli, sizeof(FILE_INFO) * fli_count);
			mem_free(&fli);
		}
		fli = tmp_fli;
		fli_size += FILELIST_ALLOC_COUNT;
	}
	// ǉ
	MoveMemory(&fli[low + 1], &fli[low], sizeof(FILE_INFO) * (fli_count - low));
	fli[low] = *new_fli;
	fli_count++;
	return TRUE;
}

/*
 * filelist_free - t@CXg̉
 */
void filelist_free(void)
{
	mem_free(&fli);
	fli_size = 0;
	fli_count = 0;
}

/*
 * string_to_time - tSYSTEMTIMEɕϊ
 */
static BOOL string_to_time(TCHAR *buf, SYSTEMTIME *tm)
{
	DATE ret_date = 0;
#ifndef UNICODE
	WCHAR wbuf[BUF_SIZE];
#endif

	ZeroMemory(tm, sizeof(SYSTEMTIME));
#ifdef UNICODE
	VarDateFromStr(buf, LANG_USER_DEFAULT, 0, &ret_date);
#else
	MultiByteToWideChar(CP_ACP, 0, buf, -1, wbuf, BUF_SIZE - 1);
	VarDateFromStr(wbuf, LANG_USER_DEFAULT, 0, &ret_date);
#endif
	VariantTimeToSystemTime(ret_date, tm);
	return TRUE;
}

/*
 * get_statistics - f[^ڎ擾
 */
static void get_statistics(const int type, DATA_INFO *di, const TCHAR *title, const TCHAR *st)
{
	static UINT hash[29];
	UINT t_hash;

	if (hash[0] == 0) {
		hash[0] = str2hash(TEXT("Sets"));
		hash[1] = str2hash(TEXT("Legs"));
		hash[2] = str2hash(TEXT("Tons"));
		hash[3] = str2hash(TEXT("100+"));
		hash[4] = str2hash(TEXT("140+"));
		hash[5] = str2hash(TEXT("180's"));
		hash[6] = str2hash(TEXT("High Out"));
		hash[7] = str2hash(TEXT("High Off"));			// old
		hash[8] = str2hash(TEXT("Best"));
		hash[9] = str2hash(TEXT("Short"));				// old
		hash[10] = str2hash(TEXT("Worst"));

		hash[11] = str2hash(TEXT("2-80 Success"));
		hash[12] = str2hash(TEXT("2-80 Failure"));
		hash[13] = str2hash(TEXT("81-130 Success"));
		hash[14] = str2hash(TEXT("81-130 Failure"));
		hash[15] = str2hash(TEXT("131+ Success"));
		hash[16] = str2hash(TEXT("131+ Failure"));

		hash[17] = str2hash(TEXT("Darts Won"));
		hash[18] = str2hash(TEXT("Darts (Won)"));		// old
		hash[19] = str2hash(TEXT("Score"));
		hash[20] = str2hash(TEXT("Darts"));
		hash[21] = str2hash(TEXT("First 9 Score"));
		hash[22] = str2hash(TEXT("First 9 Darts"));
		hash[23] = str2hash(TEXT("Check-Out"));
		hash[24] = str2hash(TEXT("Double"));
		hash[25] = str2hash(TEXT("Keep Won"));
		hash[26] = str2hash(TEXT("Keep All"));
		hash[27] = str2hash(TEXT("Break Won"));
		hash[28] = str2hash(TEXT("Break All"));
#ifdef _DEBUG
		{
			int i, j;

			for (i = 0; i < sizeof(hash) / sizeof(UINT); i++) {
				for (j = 0; j < sizeof(hash) / sizeof(UINT); j++) {
					if (hash[i] == hash[j] && i != j) {
						MessageBox(NULL, TEXT("Error match: get_statistics()"), TEXT("debug"), 0);
						break;
					}
				}
			}
		}
#endif
	}

	switch (type) {
	case 0:
		t_hash = str2hash(title);
		if (t_hash == hash[0]) {di->sets = _ttoi(st); break;}
		if (t_hash == hash[1]) {di->legs = di->win_count = _ttoi(st); break;}
		if (t_hash == hash[2]) {di->ton_count = _ttoi(st); break;}
		if (t_hash == hash[3]) {di->ton00_count = _ttoi(st); break;}
		if (t_hash == hash[4]) {di->ton40_count = _ttoi(st); break;}
		if (t_hash == hash[5]) {di->ton80_count = _ttoi(st); break;}
		if (t_hash == hash[6]) {di->high_off = _ttoi(st); break;}
		if (t_hash == hash[7]) {di->high_off = _ttoi(st); break;}
		if (t_hash == hash[8]) {di->short_game = _ttoi(st); break;}
		if (t_hash == hash[9]) {di->short_game = _ttoi(st); break;}
		if (t_hash == hash[10]) {di->long_game = _ttoi(st); break;}
		break;
	case 2:
		t_hash = str2hash(title);
		if (t_hash == hash[11]) {di->success_2_80 = _ttoi(st); break;}
		if (t_hash == hash[12]) {di->failure_2_80 = _ttoi(st); break;}
		if (t_hash == hash[13]) {di->success_81_130 = _ttoi(st); break;}
		if (t_hash == hash[14]) {di->failure_81_130 = _ttoi(st); break;}
		if (t_hash == hash[15]) {di->success_131 = _ttoi(st); break;}
		if (t_hash == hash[16]) {di->failure_131 = _ttoi(st); break;}
		break;
	case 3:
		t_hash = str2hash(title);
		if (t_hash == hash[17]) {di->win_darts = _ttoi(st); break;}
		if (t_hash == hash[18]) {di->win_darts = _ttoi(st); break;}
		if (t_hash == hash[19]) {di->all_score = _ttoi(st); break;}
		if (t_hash == hash[20]) {di->all_darts = _ttoi(st); break;}
		if (t_hash == hash[21]) {di->first9_score = _ttoi(st); break;}
		if (t_hash == hash[22]) {di->first9_darts = _ttoi(st); break;}
		if (t_hash == hash[23]) {di->check_out = _ttoi(st); break;}
		if (t_hash == hash[24]) {di->check_out_aim = _ttoi(st); break;}
		if (t_hash == hash[25]) {di->win_keep_legs = _ttoi(st); break;}
		if (t_hash == hash[26]) {di->all_keep_legs = _ttoi(st); break;}
		if (t_hash == hash[27]) {di->win_break_legs = _ttoi(st); break;}
		if (t_hash == hash[28]) {di->all_break_legs = _ttoi(st); break;}
		break;
	}
}

/*
 * csv_get_next - ̃f[^Ɉړ
 */
static TCHAR *csv_get_next(TCHAR *buf)
{
	TCHAR *p = buf;

	if (*p == TEXT('"')) {
		for (p++; *p != TEXT('\0') && *p != TEXT('\r') && *p != TEXT('\n'); p++) {
			if (*p == TEXT('"')) {
				if (*(p + 1) != TEXT('"')) {
					break;
				}
				p++;
			}
		}
		return p;
	}
	for (; *p != TEXT('\0') && *p != TEXT('\r') && *p != TEXT('\n') && *p != TEXT(','); p++);
	return p;
}

/*
 * csv_del_qoute - ڂu"v
 */
static TCHAR *csv_del_qoute(TCHAR *buf)
{
	TCHAR *p;

	for (p = buf; *p != TEXT('\0'); p++) {
		if (*p == TEXT('"') && *(p + 1) == TEXT('"')) {
			lstrcpy(p, p + 1);
		}
	}
	return buf;
}

/*
 * csv_get_data - CSVf[^擾
 */
static void csv_get_data(DATA_INFO *di, TCHAR *buf)
{
	TCHAR *p = buf;
	TCHAR *r;
	TCHAR tmp[BUF_SIZE];
	int type = 0;
	int i;

	// t擾
	r = csv_get_next(p);
	if (*p == TEXT('"')) p++;
	lstrcpyn(tmp, p, r - p + 1);
	string_to_time(csv_del_qoute(tmp), &di[0].start_time);
	di[1].start_time = di[0].start_time;
	p = r;
	if (*p == TEXT('"')) p = csv_get_next(p + 1);

	// vC[擾
	if (*p == TEXT(',')) {
		for (; *p == TEXT(','); p++);
		r = csv_get_next(p);
		if (*p == TEXT('"')) p++;
		lstrcpyn(tmp, p, r - p + 1);
		lstrcpyn(di[0].name, csv_del_qoute(tmp), NAME_SIZE - 1);
		di[0].name_hash = str2hash(di[0].name);
		p = r;
		if (*p == TEXT('"')) p = csv_get_next(p + 1);
		if (*p == TEXT(',')) {
			for (; *p == TEXT(','); p++);
			r = csv_get_next(p);
			if (*p == TEXT('"')) p++;
			lstrcpyn(tmp, p, r - p + 1);
			lstrcpyn(di[1].name, csv_del_qoute(tmp), NAME_SIZE - 1);
			di[1].name_hash = str2hash(di[1].name);
			p = r;
			if (*p == TEXT('"')) p = csv_get_next(p + 1);
		} else {
			// error
		}
	} else {
			// error
	}
	for (; *p != TEXT('\0') && *p != TEXT('\r') && *p != TEXT('\n'); p++);

	// v擾
	while (*p != TEXT('\0') && str_cmp_ni(p, TEXT("\r\n\r\n"), 4) != 0) {
		for (; *p == TEXT('\r') || *p == TEXT('\n'); p++);
		// v^Cv擾
		switch (*p) {
		case TEXT('T'):
		case TEXT('t'):
			// Totals
			type = 0;
			break;
		case TEXT('A'):
		case TEXT('a'):
			// Averages
			type = 1;
			break;
		case TEXT('F'):
		case TEXT('f'):
			// Finish Stats
			type = 2;
			break;
		case TEXT('C'):
		case TEXT('c'):
			// Counts
			type = 3;
			break;
		}
		for (i = 0; i < 2; i++) {
			p = csv_get_next(p);
			if (*p == TEXT(',')) {
				p++;
				r = csv_get_next(p);
				lstrcpyn(tmp, p, r - p + 1);
				p = r;
				if (*p == TEXT(',')) {
					p++;
					// v擾
					get_statistics(type, &di[i], tmp, p);
				}
			}
		}
		for (; *p != TEXT('\0') && *p != TEXT('\r') && *p != TEXT('\n'); p++);
	}
	for (; *p == TEXT('\r') || *p == TEXT('\n'); p++);

	// XRA擾
	while (*p != TEXT('\0')) {
		if (*p == TEXT(',')) {
			// JnXRA擾
			for (; *p == TEXT(',') || *p == TEXT('*'); p++);
			di[0].start_score = _ttoi(p);
			for (; *p != TEXT('\0') && *p != TEXT('\r') && *p != TEXT('\n') && *p != TEXT(','); p++);
			for (; *p == TEXT(',') || *p == TEXT('*'); p++);
			di[1].start_score = _ttoi(p);
			break;
		}
		for (; *p != TEXT('\0') && *p != TEXT('\r') && *p != TEXT('\n'); p++);
		for (; *p == TEXT('\r') || *p == TEXT('\n'); p++);
	}
}

/*
 * csv_read_file - CSVt@Cǂݍ
 */
int csv_read_file(DATA_INFO *di, const TCHAR *path, const TCHAR *filename, const int index)
{
	TCHAR file_path[MAX_PATH];
	TCHAR err_str[BUF_SIZE];
	TCHAR *buf;
	char *cbuf;
	DWORD size;
#ifdef UNICODE
	DWORD len;
#endif	//UNICODE

	// t@CJ
	lstrcpy(file_path, path);
	lstrcat(file_path, TEXT("\\"));
	lstrcat(file_path, filename);
	cbuf = file_read_buf(file_path, &size, err_str);
	if (cbuf == NULL) {
		return -1;
	}
	*(cbuf + size) = '\0';
#ifdef UNICODE
	len = MultiByteToWideChar(CP_ACP, 0, cbuf, size, NULL, 0);
	if ((buf = (TCHAR *)mem_alloc(sizeof(TCHAR) * (len + 1))) == NULL) {
		mem_free(&cbuf);
		return -1;
	}
	MultiByteToWideChar(CP_ACP, 0, cbuf, size, buf, len);
	size = len;
	mem_free(&cbuf);
#else	//UNICODE
	buf = cbuf;
#endif	//UNICODE
	lstrcpyn(di[0].file_name, filename, FILE_NAME_SIZE - 1);
	lstrcpyn(di[1].file_name, filename, FILE_NAME_SIZE - 1);
	di[0].file_name_hash = di[1].file_name_hash = str2hash(filename);
	csv_get_data(di, buf);
	di[0].status = di[1].status = DATA_STATUS_NONE;
	di[0].index = di[1].index = -1;
	di[0].vs_index = index + 1;
	di[1].vs_index = index;
	di[0].sets_play = di[1].sets_play = 1;
	di[0].legs_play = di[1].legs_play = di[0].legs + di[1].legs;
	mem_free(&buf);
	return 2;
}

/*
 * csv_read - CSVt@C1ǂݍ
 */
BOOL csv_read(const TCHAR *path, const TCHAR *file, DATA_INFO **di, int *di_count, int *di_size)
{
	DATA_INFO *tmp_di;
	FILE_INFO tmp_fli;
	int cnt;
	int index;
	int i;

	tmp_fli.file_name_hash = str2hash(file);
	lstrcpyn(tmp_fli.file_name, file, FILE_NAME_SIZE - 1);
	// ɒǉĂt@C`FbN
	if (filelist_add(&tmp_fli) == FALSE) {
		return TRUE;
	}
	if (*di_count + 2 > *di_size) {
		// Ċm
		tmp_di = mem_calloc(sizeof(DATA_INFO) * ((*di_count) + 2 + DATA_ALLOC_SIZE));
		if (tmp_di == NULL) {
			return FALSE;
		}
		if (*di != NULL) {
			CopyMemory(tmp_di, *di, sizeof(DATA_INFO) * (*di_count));
			mem_free(&(*di));
		}
		*di = tmp_di;
		*di_size = (*di_count) + 2 + DATA_ALLOC_SIZE;
	}
	cnt = csv_read_file(&(*di)[*di_count], path, file, *di_count);
	if (cnt == -1) {
		return FALSE;
	}
	index = *di_count;
	*di_count += cnt;
	if (op.auto_split == 1) {
		// 
		for (i = 0; i < cnt; i++) {
			data_split_data(NULL, op.auto_split_sep, op.auto_split_delete, di, di_count, di_size, index + i);
		}
	}
	return TRUE;
}

/*
 * csv_read_sync - CSV𓯊ēǂݍ
 */
BOOL csv_read_sync(const HWND hWnd, const TCHAR *path, DATA_INFO **di, int *di_count, int *di_size, int *cnt)
{
	WIN32_FIND_DATA find_data;
	HANDLE h_find_file;
	TCHAR find_path[MAX_PATH];
	BOOL ret = TRUE;
	int i;

	// CSV̓ǂݍ
	wsprintf(find_path, TEXT("%s\\*"), path);
	if ((h_find_file = FindFirstFile(find_path, &find_data)) == INVALID_HANDLE_VALUE) {
		return ret;
	}
	do {
		if (find_data.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY) {
			if (lstrcmp(find_data.cFileName, TEXT(".")) != 0 && lstrcmp(find_data.cFileName, TEXT("..")) != 0) {
				wsprintf(find_path, TEXT("%s\\%s"), path, find_data.cFileName);
				if (csv_read_sync(hWnd, find_path, di, di_count, di_size, cnt) == FALSE) {
					ret = FALSE;
					break;
				}
			}
			continue;
		}
		if (str_match(TEXT("*.csv"), find_data.cFileName) == FALSE) {
			continue;
		}
		i = *di_count;
		csv_read(path, find_data.cFileName, di, di_count, di_size);
		if (cnt != NULL && i != *di_count) {
			(*cnt)--;
			if (*cnt < 0) {
				(*cnt) = FILE_READ_COUNT;
				SendMessage(hWnd, WM_MAIN_RELOAD_STATUS, 0, 0);
				do_events(hWnd);
				if (IsWindow(hWnd) == FALSE || cancel_flag == TRUE) {
					cancel_flag = FALSE;
					ret = FALSE;
					break;
				}
			}
		}
	} while (FindNextFile(h_find_file, &find_data) == TRUE);
	FindClose(h_find_file);
	return ret;
}

/*
 * csv_read_cancel - CSVt@C̃LZ
 */
void csv_read_cancel(const BOOL flag)
{
	cancel_flag = flag;
}

/*
 * csv_find_index - CSVt@CɊ֘ACfbNX
 */
int csv_find_index(const TCHAR *file, DATA_INFO *di, int di_count)
{
	TCHAR file_name[FILE_NAME_SIZE];
	UINT file_name_hash;
	int i;

	file_name_hash = str2hash(file);
	lstrcpyn(file_name, file, FILE_NAME_SIZE - 1);
	for (i = 0; i < di_count; i++) {
		if (file_name_hash == di[i].file_name_hash && lstrcmpi(file_name, di[i].file_name) == 0) {
			break;
		}
	}
	if (i >= di_count) {
		return -1;
	}
	return i;
}

/*
 * csv_count - CSVt@C̃JEg
 */
int csv_count(const TCHAR *path)
{
	WIN32_FIND_DATA find_data;
	HANDLE h_find_file;
	TCHAR find_path[MAX_PATH];
	int count = 0;

	// t@C̃JEg
	wsprintf(find_path, TEXT("%s\\*"), path);
	if ((h_find_file = FindFirstFile(find_path, &find_data)) == INVALID_HANDLE_VALUE) {
		return count;
	}
	do {
		if (find_data.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY) {
			if (lstrcmp(find_data.cFileName, TEXT(".")) != 0 && lstrcmp(find_data.cFileName, TEXT("..")) != 0) {
				wsprintf(find_path, TEXT("%s\\%s"), path, find_data.cFileName);
				count += csv_count(find_path);
			}
			continue;
		}
		if (str_match(TEXT("*.csv"), find_data.cFileName) == FALSE) {
			continue;
		}
		count++;
	} while (FindNextFile(h_find_file, &find_data) == TRUE);
	FindClose(h_find_file);
	return count;
}

/*
 * csv_score_read - CSVt@CXRAǂݍ
 */
BOOL csv_score_read(const TCHAR *file_path, CSV_SCORE *csvs)
{
	CSV_SCORE_INFO tmp_csvsi[2];
	CSV_SCORE_INFO *alloc_csvsi;
	TYPE_SCORE *tmp_score;
	int *tmp_left;
	TCHAR err_str[BUF_SIZE];
	TCHAR *buf;
	TCHAR *p, *r;
	char *cbuf;
	long file_size;
	int round;
	int score_alloc = DATA_ALLOC_SIZE;
	int i;
#ifdef UNICODE
	long len;
#endif

	// t@Cǂݍ
	cbuf = file_read_buf(file_path, &file_size, err_str);
	if (cbuf == NULL) {
		return FALSE;
	}
	*(cbuf + file_size) = '\0';

#ifdef UNICODE
	len = MultiByteToWideChar(CP_ACP, 0, cbuf, -1, NULL, 0);
	if ((buf = (TCHAR *)mem_alloc(sizeof(TCHAR) * (len + 1))) == NULL) {
		mem_free(&cbuf);
		return FALSE;
	}
	MultiByteToWideChar(CP_ACP, 0, cbuf, -1, buf, len);
	file_size = len;
	mem_free(&cbuf);
#else
	buf = cbuf;
#endif

	// vXLbv
	for (p = buf; (file_size > (p - buf)) && *p != TEXT('\0') &&
		CompareString(MAKELCID(MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT), SORT_DEFAULT),
		0, p, 4, TEXT("\r\n\r\n"), 4) != 2; p++);
	if ((file_size <= (p - buf)) || *p == TEXT('\0')) {
		mem_free(&cbuf);
		return TRUE;
	}
	p += 4;

	ZeroMemory(tmp_csvsi, sizeof(CSV_SCORE_INFO) * 2);
	tmp_csvsi[0].score = mem_calloc(sizeof(TYPE_SCORE) * score_alloc);
	tmp_csvsi[0].left = mem_calloc(sizeof(int) * score_alloc);
	tmp_csvsi[1].score = mem_calloc(sizeof(TYPE_SCORE) * score_alloc);
	tmp_csvsi[1].left = mem_calloc(sizeof(int) * score_alloc);
	// XRA̓ǂݍ
	while ((file_size > (p - buf)) && *p != TEXT('\0')) {
		for (r = p; (file_size > (r - buf)) && (*r != TEXT('\r') && *r != TEXT('\n')); r++);

		if (*p == TEXT(',')) {
			// JnXRA擾
			for (; *p == TEXT(','); p++);
			if (*p == TEXT('*')) {
				tmp_csvsi[0].first = tmp_csvsi[1].first = 0;
			} else {
				tmp_csvsi[0].first = tmp_csvsi[1].first = 1;
			}
			for (; *p == TEXT(',') || *p == TEXT('*'); p++);
			tmp_csvsi[0].start_score = _ttoi(p);
			for (; *p != TEXT('\0') && *p != TEXT('\r') && *p != TEXT('\n') && *p != TEXT(','); p++);
			for (; *p == TEXT(',') || *p == TEXT('*'); p++);
			tmp_csvsi[1].start_score = _ttoi(p);

		} else if (*p >= TEXT('0') && *p <= TEXT('9')) {
			// ̓XRA擾
			round = _ttoi(p) - 1;
			if (round >= 0) {
				if (round >= score_alloc) {
					for (i = 0; i < 2; i++) {
						tmp_score = mem_calloc(sizeof(TYPE_SCORE) * (round + DATA_ALLOC_SIZE));
						CopyMemory(tmp_score, tmp_csvsi[i].score, sizeof(TYPE_SCORE) * score_alloc);
						mem_free(&tmp_csvsi[i].score);
						tmp_csvsi[i].score = tmp_score;

						tmp_left= mem_calloc(sizeof(int) * (round + DATA_ALLOC_SIZE));
						CopyMemory(tmp_left, tmp_csvsi[i].left, sizeof(int) * score_alloc);
						mem_free(&tmp_csvsi[i].left);
						tmp_csvsi[i].left = tmp_left;
					}
					score_alloc = round + DATA_ALLOC_SIZE;
				}
				// Round
				for (; *p != TEXT('\0') && *p != TEXT('\r') && *p != TEXT('\n') && *p != TEXT(','); p++);
				if (*p == TEXT(',')) {
					p++;
				}
				// Scored
				if ((*p >= TEXT('0') && *p <= TEXT('9')) || *p == TEXT('-')) {
					tmp_csvsi[0].score[round] = _ttoi(p);
				} else if (*p == TEXT('(')) {
					// o[W
					p++;
					tmp_csvsi[0].score[round] = _ttoi(p) * -1;
				}
				if (tmp_csvsi[0].score[round] < 0) {
					// finish
					tmp_csvsi[0].player = tmp_csvsi[1].player = 1;
					tmp_csvsi[0].round = tmp_csvsi[1].round = round;
					if (tmp_csvsi[0].first == 1) {
						tmp_csvsi[0].round++;
						tmp_csvsi[1].round++;
					}
				}
				for (; *p != TEXT('\0') && *p != TEXT('\r') && *p != TEXT('\n') && *p != TEXT(','); p++);
				if (*p == TEXT(',')) {
					p++;
				}
				// To Go
				tmp_csvsi[0].left[round] = _ttoi(p);
				for (; *p != TEXT('\0') && *p != TEXT('\r') && *p != TEXT('\n') && *p != TEXT(','); p++);
				if (*p == TEXT(',')) {
					p++;
				}
				// Scored
				if ((*p >= TEXT('0') && *p <= TEXT('9')) || *p == TEXT('-')) {
					tmp_csvsi[1].score[round] = _ttoi(p);
				} else if (*p == TEXT('(')) {
					// o[W
					p++;
					tmp_csvsi[1].score[round] = _ttoi(p) * -1;
				}
				if (tmp_csvsi[1].score[round] < 0) {
					// finish
					tmp_csvsi[0].player = tmp_csvsi[1].player = 0;
					tmp_csvsi[0].round = tmp_csvsi[1].round = round;
					if (tmp_csvsi[0].first == 0) {
						tmp_csvsi[0].round++;
						tmp_csvsi[1].round++;
					}
				}
				for (; *p != TEXT('\0') && *p != TEXT('\r') && *p != TEXT('\n') && *p != TEXT(','); p++);
				if (*p == TEXT(',')) {
					p++;
				}
				if (*p != TEXT('\r') && *p != TEXT('\n')) {
					// To Go
					tmp_csvsi[1].left[round] = _ttoi(p);
				}
			}

		} else if (*p == TEXT('D')) {
			for (; *p != TEXT('\0') && *p != TEXT('\r') && *p != TEXT('\n') && *p != TEXT(','); p++);
			if (*p == TEXT(',')) {
				p++;
			}
			if (*p != TEXT(',')) {
				tmp_csvsi[0].darts = _ttoi(p);
			} else {
				for (; *p == TEXT(','); p++);
				tmp_csvsi[1].darts = _ttoi(p);
			}
			if (tmp_csvsi[0].round == 0) {
				tmp_csvsi[0].player = tmp_csvsi[1].player = tmp_csvsi[0].first;
				tmp_csvsi[0].round = tmp_csvsi[1].round = round + 1;
			}

			// bÖړ
			if (csvs->csvsi_size < csvs->csvsi_count + 2) {
				csvs->csvsi_size += DATA_ALLOC_SIZE;
				alloc_csvsi = (CSV_SCORE_INFO *)mem_calloc(sizeof(CSV_SCORE_INFO) * csvs->csvsi_size);
				if (alloc_csvsi != NULL) {
					if (csvs->csvsi != NULL) {
						CopyMemory(alloc_csvsi, csvs->csvsi, sizeof(CSV_SCORE_INFO) * csvs->csvsi_count);
						mem_free(&csvs->csvsi);
					}
					csvs->csvsi = alloc_csvsi;
				}
			}
			csvs->csvsi[csvs->csvsi_count] = tmp_csvsi[0];
			csvs->csvsi[csvs->csvsi_count + 1] = tmp_csvsi[1];
			csvs->csvsi_count += 2;

			score_alloc = DATA_ALLOC_SIZE;
			tmp_csvsi[0].round = tmp_csvsi[1].round = 0;
			tmp_csvsi[0].player = tmp_csvsi[1].player = 0;
			tmp_csvsi[0].darts = tmp_csvsi[1].darts = 0;
			tmp_csvsi[0].first = tmp_csvsi[1].first = 0;
			tmp_csvsi[0].score = mem_calloc(sizeof(TYPE_SCORE) * score_alloc);
			tmp_csvsi[0].left = mem_calloc(sizeof(int) * score_alloc);
			tmp_csvsi[1].first = FALSE;
			tmp_csvsi[1].score = mem_calloc(sizeof(TYPE_SCORE) * score_alloc);
			tmp_csvsi[1].left = mem_calloc(sizeof(int) * score_alloc);
		}
		p = r;
		for (; (file_size > (p - buf)) && (*p == TEXT('\r') || *p == TEXT('\n')); p++);
	}
	mem_free(&tmp_csvsi[0].score);
	mem_free(&tmp_csvsi[0].left);
	mem_free(&tmp_csvsi[1].score);
	mem_free(&tmp_csvsi[1].left);
	mem_free(&buf);
	return TRUE;
}

/*
 * csv_score_free - XRẢ
 */
void csv_score_free(CSV_SCORE *csvs)
{
	int i;

	for (i = 0; i < csvs->csvsi_count; i++) {
		mem_free(&csvs->csvsi[i].score);
		mem_free(&csvs->csvsi[i].left);
	}
	mem_free(&csvs->csvsi);
}

/*
 * csv_put_conv - CSVo͗p̍쐬
 */
static TCHAR *csv_put_conv(const TCHAR *buf)
{
	static TCHAR ret[BUF_SIZE];
	TCHAR *r;
	int i;

	for (i = 0; *(buf + i) != TEXT('\0') && *(buf + i) != TEXT(',') && *(buf + i) != TEXT('"'); i++);
	if (*(buf + i) == TEXT('\0')) {
		return (TCHAR *)buf;
	}
	r = ret;
	*(r++) = TEXT('"');
	for (i = 0; *(buf + i) != TEXT('\0'); i++) {
		if (*(buf + i) == TEXT('"')) {
			*(r++) = TEXT('"');
		}
		*(r++) = *(buf + i);
	}
	*(r++) = TEXT('"');
	*r = TEXT('\0');
	return ret;
}

/*
 * csv_write_header - CSṼwb_
 */
BOOL csv_write_header(const TCHAR *path, const TCHAR *filename, const DATA_INFO *di)
{
	HANDLE hFile;
	TCHAR file_path[MAX_PATH];
	TCHAR err_str[BUF_SIZE];
	TCHAR data[BUF_SIZE];
	TCHAR tmp[BUF_SIZE];
	char *buf;
	char *p;
	DWORD size;
	DWORD ret;

	// t@C
	if (file_find(path, filename, file_path) == FALSE) {
		return FALSE;
	}
	// t@Cǂݍ
	buf = file_read_buf(file_path, &size, err_str);
	if (buf == NULL) {
		return FALSE;
	}
	// t@CJ
	hFile = CreateFile(file_path, GENERIC_READ | GENERIC_WRITE, 0, 0, CREATE_ALWAYS, FILE_ATTRIBUTE_NORMAL, NULL);
	if (hFile == NULL || hFile == (HANDLE)-1) {
		mem_free(&buf);
		return FALSE;
	}
	// wb_̏
	data_to_value(HDTP_DATE, di, tmp);
	lstrcpy(data, csv_put_conv(tmp));
	lstrcat(data, TEXT(","));
	data_to_value(HDTP_NAME, di, tmp);
	lstrcat(data, csv_put_conv(tmp));
	lstrcat(data, TEXT(",,"));
	data_to_value(HDTP_VS_NAME, di, tmp);
	lstrcat(data, csv_put_conv(tmp));
	lstrcat(data, TEXT(",\r\n"));
	if (WriteFile(hFile, data, lstrlen(data), &ret, NULL) == FALSE) {
		CloseHandle(hFile);
		mem_free(&buf);
		return FALSE;
	}
	// f[^̏
	for (p = buf; (DWORD)(p - buf) < size && *p != TEXT('\r') && *p != TEXT('\n'); p++);
	for (; (DWORD)(p - buf) < size && (*p == TEXT('\r') || *p == TEXT('\n')); p++);
	if (WriteFile(hFile, p, size - (p - buf), &ret, NULL) == FALSE) {
		CloseHandle(hFile);
		mem_free(&buf);
		return FALSE;
	}
	FlushFileBuffers(hFile);
	CloseHandle(hFile);
	mem_free(&buf);
	return TRUE;
}

/*
 * csv_write_list - ꗗCSVo
 */
BOOL csv_write_list(const HWND lv_wnd, const TCHAR *file_name, const DATA_INFO *di, const int di_count)
{
	const DATA_INFO *out_di;
	HANDLE hFile;
	TCHAR data[32768];
	TCHAR tmp[BUF_SIZE];
	TCHAR *p;
	DWORD ret;
	int i, j;

	// t@CJ
	hFile = CreateFile(file_name, GENERIC_READ | GENERIC_WRITE, 0, 0, CREATE_ALWAYS, FILE_ATTRIBUTE_NORMAL, NULL);
	if (hFile == NULL || hFile == (HANDLE)-1) {
		return FALSE;
	}
	// wb_o
	*data = TEXT('\0');
	for (j = HDTP_NAME; j <= HDTP_FAILURE_131; j++) {
		if (j != HDTP_NAME) {
			lstrcat(data, TEXT(","));
		}
		lstrcpy(data + lstrlen(data), data_get_title(j));
	}
	lstrcat(data, TEXT("\r\n"));
	if (WriteFile(hFile, data, lstrlen(data), &ret, NULL) == FALSE) {
		CloseHandle(hFile);
		return FALSE;
	}
	// ꗗo
	for (i = 0; i < ListView_GetItemCount(lv_wnd); i++) {
		out_di = &di[listview_get_lparam(lv_wnd, i)];
		*data = TEXT('\0');
		for (j = HDTP_NAME; j <= HDTP_FAILURE_131; j++) {
			if (j != HDTP_NAME) {
				lstrcat(data, TEXT(","));
			}
			if (j == HDTP_DATE || j == HDTP_NAME || j == HDTP_VS_NAME) {
				data_to_value(j, out_di, tmp);
				for (p = tmp; *p != TEXT('\0'); p++) {
					if (*p == TEXT(',') || *p == TEXT('"')) {
						*p = TEXT('_');
					}
				}
				lstrcat(data, tmp);
			} else {
				data_to_value(j, out_di, data + lstrlen(data));
			}
		}
		lstrcat(data, TEXT("\r\n"));
		if (WriteFile(hFile, data, lstrlen(data), &ret, NULL) == FALSE) {
			CloseHandle(hFile);
			return FALSE;
		}
	}
	FlushFileBuffers(hFile);
	CloseHandle(hFile);
	return TRUE;
}

/*
 * csv_copy_list_length - ꗗőIĂf[^̒̕
 */
static int csv_copy_list_length(const HWND lv_wnd, const DATA_INFO *di, const int di_count)
{
	const DATA_INFO *out_di;
	TCHAR tmp[BUF_SIZE];
	int i, j;
	int ret = 0;

	// wb_o
	for (j = HDTP_NAME; j <= HDTP_FAILURE_131; j++) {
		if (j != HDTP_NAME) {
			ret++;
		}
		ret += lstrlen(data_get_title(j));
	}
	ret += 2;

	// ꗗo
	i = -1;
	while ((i = ListView_GetNextItem(lv_wnd, i, LVNI_SELECTED)) != -1) {
		out_di = &di[listview_get_lparam(lv_wnd, i)];
		for (j = HDTP_NAME; j <= HDTP_FAILURE_131; j++) {
			if (j != HDTP_NAME) {
				ret++;
			}
			data_to_value(j, out_di, tmp);
			ret += lstrlen(tmp);
		}
		ret += 2;
	}
	return ret;
}

/*
 * csv_copy_list - ꗗőIĂf[^̕쐬
 */
HANDLE csv_copy_list(const HWND lv_wnd, const DATA_INFO *di, const int di_count)
{
	const DATA_INFO *out_di;
	HANDLE hMem;
	TCHAR tmp[BUF_SIZE];
	TCHAR *p, *r;
	int i, j;
	int len;

	len = csv_copy_list_length(lv_wnd, di, di_count);
	if ((hMem = GlobalAlloc(GHND, sizeof(TCHAR) * (len + 1))) == NULL) {
		return NULL;
	}
	if ((r = GlobalLock(hMem)) == NULL) {
		GlobalFree(hMem);
		return FALSE;
	}
	*r = TEXT('\0');

	// wb_o
	for (j = HDTP_NAME; j <= HDTP_FAILURE_131; j++) {
		if (j != HDTP_NAME) {
			*(r++) = TEXT(',');
		}
		lstrcpy(r, data_get_title(j));
		r += lstrlen(r);
	}
	*(r++) = TEXT('\r');
	*(r++) = TEXT('\n');

	// ꗗo
	i = -1;
	while ((i = ListView_GetNextItem(lv_wnd, i, LVNI_SELECTED)) != -1) {
		out_di = &di[listview_get_lparam(lv_wnd, i)];
		for (j = HDTP_NAME; j <= HDTP_FAILURE_131; j++) {
			if (j != HDTP_NAME) {
				*(r++) = TEXT(',');
			}
			if (j == HDTP_DATE || j == HDTP_NAME || j == HDTP_VS_NAME) {
				data_to_value(j, out_di, tmp);
				for (p = tmp; *p != TEXT('\0'); p++) {
					if (*p == TEXT(',') || *p == TEXT('"')) {
						*p = TEXT('_');
					}
				}
				lstrcpy(r, tmp);
			} else {
				data_to_value(j, out_di, r);
			}
			r += lstrlen(r);
		}
		*(r++) = TEXT('\r');
		*(r++) = TEXT('\n');
	}
	*r = TEXT('\0');
	GlobalUnlock(hMem);
	return hMem;
}
/* End of source */
